#include <GL/glut.h>
#include <GL/glu.h>
#include <GL/gl.h>

#include "globals.h"
#include "sizes.h"
#include "input.h"
#include "game.h"
#include "status.h"
#include "menu.h"


Blit BlitArray[MAX_BLITS];
int NumBlits;
Color White, Black, Red, Green, Blue, DarkGray;


void OverlaySprite(Sprite *sprite, unsigned long xpos,
		   unsigned long ypos, Blit *blit, int frame);


/*
  Function to queue a sprite to be displayed.  The sprite is either merged with
  an existing sprite, or it is queued in a blit array to be displayed.
*/
void QueueSprite(unsigned long xpos, unsigned long ypos,
		 Sprite *sprite, int frame)
{
  int i;
  int memsize;

  // Check if the sprite is on the screen
  if (xpos > ScreenWidth || ypos > ScreenHeight) {
    return;
  }

  /*
    Check to see if the sprite overlaps with an existing sprite, if it does,
    then overlay the sprite
  */
  for (i=0; i<NumBlits; i++) {
    if (xpos > BlitArray[i].xpos) {
      if (ypos > BlitArray[i].ypos) {
	if (((xpos - BlitArray[i].xpos) < BlitArray[i].Width) &&
	    ((ypos - BlitArray[i].ypos) < BlitArray[i].Height)) {
	  OverlaySprite(sprite, xpos, ypos, &(BlitArray[i]), frame);
	  return;
	}
      } else {
	if (((xpos - BlitArray[i].xpos) < BlitArray[i].Width) &&
	    ((BlitArray[i].ypos - ypos) < sprite->Height)) {
	  OverlaySprite(sprite, xpos, ypos, &(BlitArray[i]), frame);
	  return;
	}
      }
    } else {
      if (ypos > BlitArray[i].ypos) {
	if (((BlitArray[i].xpos - xpos) < sprite->Width) &&
	    ((ypos - BlitArray[i].ypos) < BlitArray[i].Height)) {
	  OverlaySprite(sprite, xpos, ypos, &(BlitArray[i]), frame);
	  return;
	}
      } else {
	if (((BlitArray[i].xpos - xpos) < sprite->Width) &&
	    ((BlitArray[i].ypos - ypos) < sprite->Height)) {
	  OverlaySprite(sprite, xpos, ypos, &(BlitArray[i]), frame);
	  return;
	}
      }
    }
  }

  // Add the sprite to the blit array
  BlitArray[NumBlits].xpos=xpos;
  BlitArray[NumBlits].ypos=ypos;
  BlitArray[NumBlits].Width=sprite->Width;
  BlitArray[NumBlits].Height=sprite->Height;
  memsize=sprite->Width*sprite->Height*3;
  BlitArray[NumBlits].pixels=(unsigned char *)malloc(memsize);
  memcpy(BlitArray[NumBlits].pixels, sprite->pixels[frame], memsize);
  NumBlits++;
}


// Function to draw on the screen all the queued up sprites
void DrawBlits()
{
  int i;
  float x, y;

  for (i=0; i<NumBlits; i++) {
    x=BlitArray[i].xpos;
    y=BlitArray[i].ypos;

    /*
      Change the coordinates to OpenGL format.  OpenGL, by default, sets the
      screen dimensions to be between -1.0 and 1.0 for both x and y
      coordinates.  Each x, y pair on the screen is a float between -1.0 and
      1.0.
    */
    x=(x/ScreenWidth)*2.0-1.0;
    y=(y/ScreenHeight)*2.0-1.0;
    glRasterPos2f(x, y);
    glDrawPixels(BlitArray[i].Width, BlitArray[i].Height, GL_RGB,
		 GL_UNSIGNED_BYTE, BlitArray[i].pixels);
    free(BlitArray[i].pixels);
  }

  NumBlits=0;
}


/*
  Function to overlay one sprite over another.  This function takes the
  two rectangles around each sprite, draws a bigger rectangle around both
  sprites, then copies each sprite into the rectangle.  The function also takes
  into account transparency.  The second sprite will be copied onto the
  rectangle first, then the original sprite will be copied wherever there is
  a black pixel.
*/
void OverlaySprite(Sprite *sprite, unsigned long xpos,
		   unsigned long ypos, Blit *blit, int frame)
{
  unsigned char *newpixels, *ptr1, *ptr2;
  int top, bottom, left, right;
  int newWidth, newHeight;
  int i, j, k;
  int memsize;

  // Find the dimensions of the new rectangle
  if (xpos + sprite->Width > blit->xpos + blit->Width) {
    right = xpos + sprite->Width;
  } else {
    right = blit->xpos + blit->Width;
  }
  if (ypos + sprite->Height > blit->ypos + blit->Height) {
    top = ypos + sprite->Height;
  } else {
    top = blit->ypos + blit->Height;
  }
  if (xpos < blit->xpos) {
    left = xpos;
  } else {
    left = blit->xpos;
  }
  if (ypos < blit->ypos) {
    bottom = ypos;
  } else {
    bottom = blit->ypos;
  }

  /*
    Find the height and width
    The height and width should be divisible by 4, this makes it easier to
    create the new sprite, since each row has to end on a 4 byte boundary for
    OpenGL
  */
  newHeight=top-bottom;
  top+=(4-newHeight%4)%4;
  newHeight=top-bottom;
  newWidth=right-left;
  right+=(4-newWidth%4)%4;
  newWidth=right-left;

  memsize=newWidth*newHeight*3;
  newpixels=(unsigned char *)malloc(memsize);
  memset(newpixels, 0, memsize);

  // Add the new sprite
  ptr1=blit->pixels;
  ptr2=newpixels;
  ptr2+=3*newWidth*(blit->ypos-bottom);
  for (i=0; i<blit->Height; i++) {
    ptr2+=3*(blit->xpos-left);
    for (j=0; j<blit->Width; j++) {
      for (k=0; k<3; k++) {
	*ptr2=*ptr1;
	ptr1++;
	ptr2++;
      }
    }
    ptr2+=3*(right-(blit->xpos+blit->Width));
  }
  free(blit->pixels);

  // Add the original sprite
  ptr1=sprite->pixels[frame];
  ptr2=newpixels;
  ptr2+=3*newWidth*(ypos-bottom);
  for (i=0; i<sprite->Height; i++) {
    ptr2+=3*(xpos-left);
    for (j=0; j<sprite->Width; j++) {
      // Only add it if the pixels already there is black
      if (*ptr2==0 && *(ptr2+1)==0 && *(ptr2+2)==0) {
	for (k=0; k<3; k++) {
	  *ptr2=*ptr1;
	  ptr2++;
	  ptr1++;
	}
      } else {
	ptr2+=3;
	ptr1+=3;
      }
    }
    ptr2+=3*(right-(xpos+sprite->Width));
  }

  blit->pixels=newpixels;
  blit->Width=newWidth;
  blit->Height=newHeight;
  blit->xpos=left;
  blit->ypos=bottom;
}


// Function to draw a line on the screen
void DrawLine(unsigned long x1pos, unsigned long y1pos,
	      unsigned long x2pos, unsigned long y2pos, Color LineColor)
{
  float x1, y1, x2, y2;

  x1=x1pos;
  y1=y1pos;
  x2=x2pos;
  y2=y2pos;

  x1=(x1/ScreenWidth)*2.0-1.0;
  y1=(y1/ScreenHeight)*2.0-1.0;
  x2=(x2/ScreenWidth)*2.0-1.0;
  y2=(y2/ScreenHeight)*2.0-1.0;

  glBegin(GL_LINES);
  glColor3f(LineColor.red, LineColor.green, LineColor.blue);
  glVertex2f(x1, y1);
  glVertex2f(x2, y2);
  glEnd();
}


/*
  Function to draw a point on the screen
  glDrawPixels is used instead of glVertex2f, because the points were showing
  up in front of the blits
*/
void DrawPoint(unsigned long xpos, unsigned long ypos, Color PointColor)
{
  float x, y;

  x=xpos;
  y=ypos;

  x=(x/ScreenWidth)*2.0-1.0;
  y=(y/ScreenHeight)*2.0-1.0;

  glBegin(GL_POINTS);
  glColor3f(PointColor.red, PointColor.green, PointColor.blue);
  glVertex2f(x, y);
  glEnd();
}


/*
  Function to draw text on the screen, it returns the x position after the text
  is drawn.
*/
unsigned long DrawText(unsigned long xpos, unsigned long ypos,
		       char *text, Color TextColor, int big=0)
{
  float x, y;
  float CharWidth;

  x=xpos;
  y=ypos;

  x=(x/ScreenWidth)*2.0-1.0;
  y=(y/ScreenHeight)*2.0-1.0;

  glColor3f(TextColor.red, TextColor.green, TextColor.blue);
  while (*text) {
    glRasterPos2f(x, y);
    if (big) {
#ifndef CYGWIN
      glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, *text);
      CharWidth=glutBitmapWidth(GLUT_BITMAP_HELVETICA_18, *text);
#else
      glutBitmapCharacter((void *)8, *text);
      CharWidth=glutBitmapWidth((void *)8, *text);
#endif
    } else {
#ifndef CYGWIN
      glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, *text);
      CharWidth=glutBitmapWidth(GLUT_BITMAP_HELVETICA_10, *text);
#else
      glutBitmapCharacter((void *)6, *text);
      CharWidth=glutBitmapWidth((void *)6, *text);
#endif
    }
    x+=(CharWidth/ScreenWidth)*2.0;
    xpos+=(int)CharWidth;
    text++;
  }
  return(xpos);
}


// Function to draw a rectangle on the screen
void DrawRect(unsigned long x1pos, unsigned long y1pos,
	      unsigned long x2pos, unsigned long y2pos,
	      unsigned long x3pos, unsigned long y3pos,
	      unsigned long x4pos, unsigned long y4pos,
	      Color LineColor)
{
  float x1, y1, x2, y2, x3, y3, x4, y4;

  x1=x1pos;
  y1=y1pos;
  x2=x2pos;
  y2=y2pos;
  x3=x3pos;
  y3=y3pos;
  x4=x4pos;
  y4=y4pos;

  x1=(x1/ScreenWidth)*2.0-1.0;
  y1=(y1/ScreenHeight)*2.0-1.0;
  x2=(x2/ScreenWidth)*2.0-1.0;
  y2=(y2/ScreenHeight)*2.0-1.0;
  x3=(x3/ScreenWidth)*2.0-1.0;
  y3=(y3/ScreenHeight)*2.0-1.0;
  x4=(x4/ScreenWidth)*2.0-1.0;
  y4=(y4/ScreenHeight)*2.0-1.0;

  glBegin(GL_QUADS);
  glColor3f(LineColor.red, LineColor.green, LineColor.blue);
  glVertex2f(x1, y1);
  glVertex2f(x2, y2);
  glVertex2f(x3, y3);
  glVertex2f(x4, y4);
  glEnd();
}


// Function to initialize the screen and setup keyboard callbacks
void InitScreen()
{
  GLenum type;

  MainMenuItemHighlighted=0;
  ConfMenuItemHighlighted=15;
  ScoreMenuItemHighlighted=1;
  NewScreenWidth=ScreenWidth;
  NewScreenHeight=ScreenHeight;

  type = GLUT_DEPTH | GLUT_RGB | GLUT_DOUBLE;
  glutInitDisplayMode(type);
  glutInitWindowSize(ScreenWidth, ScreenHeight);
  glutCreateWindow("Avoid The Roid");
  glutIgnoreKeyRepeat(1);
  glutSpecialFunc(MainMenuSpecialKey);
  glutKeyboardFunc(MainMenuKey);
  glutIdleFunc(NULL);
  glutDisplayFunc(MainMenu);
  SetupMainMenuNames();
}


void StartGame()
{
  glutMainLoop();
}


void ClearScreen()
{
  glClear(GL_COLOR_BUFFER_BIT);
}


void UpdateScreen()
{
  glutSwapBuffers();
}


void InitColors()
{
  White.red=1.0;
  White.green=1.0;
  White.blue=1.0;

  Black.red=0.0;
  Black.green=0.0;
  Black.blue=0.0;

  Red.red=1.0;
  Red.green=0.0;
  Red.blue=0.0;

  Green.red=0.0;
  Green.green=1.0;
  Green.blue=0.0;

  Blue.red=0.0;
  Blue.green=0.0;
  Blue.blue=1.0;

  DarkGray.red=0.2;
  DarkGray.green=0.2;
  DarkGray.blue=0.2;
}
